import numpy as np
import pytest

from bmi_tester.api import check_unit_is_valid


def test_get_var_itemsize(initialized_bmi, var_name):
    """Test getting a variable's itemsize"""
    itemsize = initialized_bmi.get_var_itemsize(var_name)
    assert itemsize > 0


# @pytest.mark.dependency()
def test_get_var_nbytes(initialized_bmi, var_name):
    """Test getting a variable's nbytes"""
    nbytes = initialized_bmi.get_var_nbytes(var_name)
    assert nbytes > 0


# @pytest.mark.dependency()
def test_get_var_location(initialized_bmi, var_name):
    """Test getting a variable's grid location"""
    location = initialized_bmi.get_var_location(var_name)
    assert isinstance(location, str)
    assert location in ("node", "edge", "face", "none")


# @pytest.mark.dependency(depends=["test_get_var_location"])
def test_var_on_grid(initialized_bmi, var_name):
    loc = initialized_bmi.get_var_location(var_name)
    if initialized_bmi.get_var_location(var_name) == "none":
        pytest.skip(f"var, {var_name}, is not located on a grid")

    gid = initialized_bmi.get_var_grid(var_name)
    if initialized_bmi.get_grid_type(gid) == "unstructured":
        if loc == "node":
            assert initialized_bmi.get_grid_node_count(gid) > 0
        elif loc == "edge":
            assert initialized_bmi.get_grid_edge_count(gid) > 0
        elif loc == "face":
            assert initialized_bmi.get_grid_face_count(gid) > 0


def test_get_var_type(initialized_bmi, var_name):
    """Test getting a variable's data type"""
    dtype = initialized_bmi.get_var_type(var_name)
    assert isinstance(dtype, str)

    try:
        np.empty(1, dtype=dtype)
    except TypeError:
        raise AssertionError(
            "get_var_type: bad data type name ({dtype})".format(dtype=dtype)
        )


def test_get_var_units(initialized_bmi, var_name):
    """Test the units of the variables."""
    units = initialized_bmi.get_var_units(var_name)
    assert isinstance(units, str)
    assert check_unit_is_valid(units)



from distutils.version import StrictVersion

import numpy as np
import pytest

from bmi_tester.api import empty_var_buffer

from ..conftest import BMI_VERSION_STRING, INPUT_FILE, Bmi

# from pytest_dependency import depends


BMI_VERSION = StrictVersion(BMI_VERSION_STRING)

BAD_VALUE = {"f": np.nan, "i": -999, "u": 0}


# @pytest.mark.dependency()
# def test_get_var_location(var_name):
#     if var_name == "channel_entrance_water_sediment~bedload__mass_flow_rate":
#         assert False


# @pytest.mark.dependency()
def test_get_var_location(initialized_bmi, var_name):
    """Test for get_var_location"""
    # assert False
    if BMI_VERSION < "1.1":
        pytest.skip(
            "testing BMIv{ver}: get_var_location was introduced in BMIv1.1".format(
                ver=BMI_VERSION
            )
        )

    assert hasattr(initialized_bmi, "get_var_location")

    loc = initialized_bmi.get_var_location(var_name)

    assert isinstance(loc, str)
    assert loc in ("node", "edge", "face", "none")


@pytest.mark.skip("too dangerous")
# @pytest.mark.dependency(depends=["initialize_works", "test_get_var_grid", "get_var_nbytes"], scope="session")
def test_set_input_values(staged_tmpdir, in_var_name):
    """Input values are numpy arrays."""
    with staged_tmpdir.as_cwd():
        bmi = Bmi()
        bmi.initialize(INPUT_FILE)

        values = empty_var_buffer(bmi, in_var_name)
        # values.fill(BAD_VALUE[values.dtype.kind])
        bmi.set_value(in_var_name, values)

        # if np.isnan(BAD_VALUE[values.dtype.kind]):
        #     assert np.all(np.isnan(values))
        # else:
        #     assert np.all(values == BAD_VALUE[values.dtype.kind])


# @pytest.mark.dependency(depends=["initialize_works", "test_get_var_grid", "get_var_nbytes"], scope="session")
# @pytest.mark.dependency(depends=["initialize_works"], scope="session")
# @pytest.mark.dependency(depends=["test_get_var_location"])
def test_get_output_values(request, initialized_bmi, out_var_name):
    """Output values are numpy arrays."""
    # name = "../../../../../../../Users/huttone/git/csdms/bmi-tester/bmi_tester/tests_pytest/test_var.py::test_get_var_grid[%s]" % out_var_name
    # depends(request, ["test_get_var_location[%s]" % out_var_name])

    # depends(request, [name])
    # depends(request, ["test_var.py::test_get_var_grid[%s]" % out_var_name]) # , "test_get_var_nbytes[%s]" % out_var_name], scope="session")

    # depends(initialized_bmi, ["test_get_var_grid[%s]" % out_var_name, "test_get_var_nbytes" % out_var_name], scope="session")

    values = empty_var_buffer(initialized_bmi, out_var_name)
    # values.fill(BAD_VALUE[values.dtype.kind])
    initial = values.tobytes()
    initialized_bmi.get_value(out_var_name, values)

    assert initial != values.tobytes()
    # assert np.any(values != initial)



import pkg_resources

from .api import (
    check_bmi,
    check_unit_is_dimensionless,
    check_unit_is_time,
    check_unit_is_valid,
)

__version__ = pkg_resources.get_distribution("bmi_tester").version
__all__ = [
    "check_bmi",
    "check_unit_is_valid",
    "check_unit_is_time",
    "check_unit_is_dimensionless",
]
del pkg_resources



from distutils.version import StrictVersion

import numpy as np
import pytest

from ..conftest import BMI_VERSION, skip_if_grid_type_is_not

VALID_GRID_TYPES = (
    "none",
    "scalar",
    "vector",
    "unstructured",
    "unstructured_triangular",
    "rectilinear",
    "structured_quadrilateral",
    "uniform_rectilinear",
    "uniform_rectilinear_grid",
)


# @pytest.mark.dependency()
def test_get_grid_rank(initialized_bmi, gid):
    "Test grid rank for grid {gid}".format(gid=gid)
    rank = initialized_bmi.get_grid_rank(gid)
    assert isinstance(rank, int)
    assert rank <= 3
    if initialized_bmi.get_grid_type(gid) == "scalar":
        assert rank == 0


def test_get_grid_size(initialized_bmi, gid):
    "Test grid size for grid {gid}".format(gid=gid)
    size = initialized_bmi.get_grid_size(gid)
    assert isinstance(size, int)
    assert size > 0


# @pytest.mark.dependency()
def test_get_grid_type(initialized_bmi, gid):
    "Test grid is known for grid {gid}".format(gid=gid)
    gtype = initialized_bmi.get_grid_type(gid)
    assert isinstance(gtype, str)
    assert gtype in VALID_GRID_TYPES


@pytest.mark.skipif(
    BMI_VERSION < StrictVersion("2.0"), reason="get_grid_node_count is BMI 2.0"
)
# @pytest.mark.dependency()
def test_get_grid_node_count(initialized_bmi, gid):
    "Test number of nodes in grid {gid}".format(gid=gid)
    skip_if_grid_type_is_not(initialized_bmi, gid, "unstructured")

    n_nodes = initialized_bmi.get_grid_node_count(gid)
    assert isinstance(n_nodes, int)
    assert n_nodes > 0


# @pytest.mark.dependency()
def test_get_grid_edge_count(initialized_bmi, gid):
    "Test number of edges in grid {gid}".format(gid=gid)
    skip_if_grid_type_is_not(initialized_bmi, gid, "unstructured")

    n_edges = initialized_bmi.get_grid_edge_count(gid)
    assert isinstance(n_edges, int)
    assert n_edges >= 0
    if n_edges == 0:
        assert initialized_bmi.get_grid_face_count(gid) == 0


# @pytest.mark.dependency()
def test_get_grid_face_count(initialized_bmi, gid):
    "Test number of faces in grid {gid}".format(gid=gid)
    skip_if_grid_type_is_not(initialized_bmi, gid, "unstructured")

    n_faces = initialized_bmi.get_grid_face_count(gid)
    assert isinstance(n_faces, int)
    assert n_faces >= 0


# @pytest.mark.dependency(depends=["test_get_grid_node_count", "test_get_grid_edge_count"])
def test_get_grid_edge_nodes(initialized_bmi, gid):
    "Test nodes at edges for grid {gid}".format(gid=gid)
    skip_if_grid_type_is_not(initialized_bmi, gid, "unstructured")

    n_edges = initialized_bmi.get_grid_edge_count(gid)
    n_nodes = initialized_bmi.get_grid_node_count(gid)

    if n_edges == 0:
        pytest.skip("grid has no edges")

    edge_nodes = np.full((n_edges, 2), -10, dtype=np.int32).reshape(-1)

    rtn = initialized_bmi.get_grid_edge_nodes(gid, edge_nodes)
    assert rtn is edge_nodes
    assert np.all(edge_nodes >= 0)
    assert np.all(edge_nodes < n_nodes)


@pytest.mark.skip("edges_per_face")
# @pytest.mark.dependency(depends=["test_get_grid_node_count", "test_get_grid_edge_count", "test_get_grid_face_count"])
def test_get_grid_edges_per_face(initialized_bmi, gid):
    "Test number of edges at each face for grid {gid}".format(gid=gid)
    skip_if_grid_type_is_not(initialized_bmi, gid, "unstructured")

    n_edges = initialized_bmi.get_grid_edge_count(gid)
    n_faces = initialized_bmi.get_grid_face_count(gid)

    if n_faces == 0:
        pytest.skip("grid has no faces")

    edges_per_face = np.full(n_faces, -1, dtype=np.int32)

    rtn = initialized_bmi.get_grid_edges_per_face(gid, edges_per_face)
    assert rtn is edges_per_face
    assert np.all(edges_per_face >= 3)
    assert np.all(edges_per_face < n_edges)


@pytest.mark.skip("face_edges")
@pytest.mark.dependency(
    depends=[
        "test_get_grid_node_count",
        "test_get_grid_edge_count",
        "test_get_grid_face_count",
        "test_get_grid_edges_per_face",
    ]
)
def test_get_grid_face_edges(initialized_bmi, gid):
    "Test edges at face for grid {gid}".format(gid=gid)
    skip_if_grid_type_is_not(initialized_bmi, gid, "unstructured")

    n_faces = initialized_bmi.get_grid_face_count(gid)
    n_edges = initialized_bmi.get_grid_edge_count(gid)

    if n_faces == 0:
        pytest.skip("grid has no edges")

    edges_per_face = np.full(n_faces, -1, dtype=np.int32)
    initialized_bmi.get_grid_edges_per_face(gid, edges_per_face)

    face_edges = np.full(edges_per_face.sum(), -1, dtype=np.int32)

    rtn = initialized_bmi.get_grid_face_edges(gid, face_edges)
    assert rtn is face_edges
    assert np.all(face_edges >= 0)
    assert np.all(face_edges < n_edges)



#! /usr/bin/env python
# try:
#     from pymt import cfunits
# except (FileNotFoundError, ModuleNotFoundError):
#     cfunits = None
import pytest
from pytest import approx

from bmi_tester.api import (
    check_unit_is_dimensionless,
    check_unit_is_time,
    check_unit_is_valid,
)


@pytest.mark.dependency()
def test_get_start_time(initialized_bmi):
    """Test that there is a start time."""
    start = initialized_bmi.get_start_time()

    assert isinstance(start, float)
    assert start == approx(0.0)


@pytest.mark.dependency()
def test_get_time_step(initialized_bmi):
    """Test that there is a time step."""
    time_step = initialized_bmi.get_time_step()

    assert isinstance(time_step, float)


def test_time_units_is_str(initialized_bmi):
    """Test the units of time is a str."""
    units = initialized_bmi.get_time_units()

    assert isinstance(units, str)


# @pytest.mark.skipif(cfunits is None, reason="cfunits is broken on this platform")
def test_time_units_is_valid(initialized_bmi):
    """Test the units of time are valid."""
    units = initialized_bmi.get_time_units()
    assert check_unit_is_valid(units)
    assert check_unit_is_time(units) or check_unit_is_dimensionless(units)


@pytest.mark.dependency(depends=["test_get_start_time", "test_get_end_time"])
def test_get_current_time(initialized_bmi):
    """Test that there is a current time."""
    start = initialized_bmi.get_start_time()
    now = initialized_bmi.get_current_time()
    stop = initialized_bmi.get_end_time()

    assert isinstance(now, (int, float))
    assert now <= stop
    assert now >= start


@pytest.mark.skip()
@pytest.mark.dependency(depends=["test_get_start_time"])
def test_get_end_time(initialized_bmi):
    """Test that there is a stop time."""
    start = initialized_bmi.get_start_time()
    stop = initialized_bmi.get_end_time()

    assert isinstance(stop, (int, float))
    assert stop >= start



import importlib
import os
from distutils.version import StrictVersion

import pytest
from model_metadata.scripting import as_cwd, cp


def skip_if_grid_type_is_not(bmi, gid, gtype):
    if isinstance(gtype, str):
        gtype = (gtype,)
    if bmi.get_grid_type(gid) not in gtype:
        gtypes = ", ".join(gtype)
        if len(gtype) > 1:
            pytest.skip(f"grid {gid} is not one of {gtypes}")
        else:
            pytest.skip(f"grid {gid} is not {gtypes}")


def skip_if_grid_type_is(bmi, gid, gtype):
    if isinstance(gtype, str):
        gtype = (gtype,)
    if bmi.get_grid_type(gid) in gtype:
        gtypes = ", ".join(gtype)
        if len(gtype) > 1:
            pytest.skip(f"grid {gid} is one of {gtypes}")
        else:
            pytest.skip(f"grid {gid} is {gtypes}")


def load_component(entry_point):
    module_name, cls_name = entry_point.split(":")

    component = None
    try:
        module = importlib.import_module(module_name)
    except ImportError:
        raise
    else:
        try:
            component = module.__dict__[cls_name]
        except KeyError:
            raise ImportError(cls_name)

    return component


try:
    class_to_test = os.environ["BMITEST_CLASS"]
except KeyError:
    Bmi = None
else:
    Bmi = load_component(class_to_test)
INPUT_FILE = os.environ.get("BMITEST_INPUT_FILE", None)
# BMI_VERSION_STRING = os.environ.get("BMI_VERSION_STRING", "1.1")
BMI_VERSION_STRING = os.environ.get("BMI_VERSION_STRING", "2.0")
BMI_VERSION = StrictVersion(BMI_VERSION_STRING)


def all_grids(bmi, gtype=None):
    in_names = set(bmi.get_input_var_names())
    out_names = set(bmi.get_output_var_names())

    grids = set()
    for name in in_names | out_names:
        if bmi.get_var_location(name) != "none":
            gid = bmi.get_var_grid(name)
            if gtype == bmi.get_grid_type(gid)[1] or gtype is None:
                grids.add(gid)
    return grids


count = 1


def get_test_parameters(infile=None, count=0):
    infile = infile or INPUT_FILE

    try:
        with open(".ROOT_DIR", "r") as fp:
            root_dir = fp.read()
    except IOError:
        root_dir = "."
    if count > 1:
        raise RuntimeError()
    count += 1

    bmi = Bmi()
    with as_cwd(root_dir):
        bmi.initialize(infile)

        in_names = set(bmi.get_input_var_names())
        out_names = set(bmi.get_output_var_names())
        grids = set(all_grids(bmi))

        # in_names = set()
        # out_names = set()
        # grids = set()

        bmi.finalize()

    meta = {
        "gid": grids,
        "var_name": in_names | out_names,
        "in_var_name": in_names - out_names,
        "out_var_name": out_names,
    }

    return meta


@pytest.fixture(scope="session")
def bmi():
    # return None
    return Bmi()


@pytest.fixture(scope="session")
def initialized_bmi(tmpdir_factory, infile=None, manifest=None):
    infile = os.environ.get("BMITEST_INPUT_FILE", None)
    manifest = os.environ.get("BMITEST_MANIFEST", infile or "").splitlines()

    tmp = tmpdir_factory.mktemp("data")
    with tmp.as_cwd() as prev:
        for file_ in [fname.strip() for fname in manifest]:
            if file_:
                cp(os.path.join(str(prev), file_), tmp / file_, create_dirs=True)

        bmi = Bmi()
        bmi.initialize(infile)
    # return None
    return bmi


@pytest.fixture(scope="function")
def staged_tmpdir(tmpdir, infile=None, manifest=None):
    infile = os.environ.get("BMITEST_INPUT_FILE", None)
    manifest = os.environ.get("BMITEST_MANIFEST", infile or "").splitlines()
    with tmpdir.as_cwd() as prev:
        for file_ in [fname.strip() for fname in manifest]:
            if file_:
                cp(os.path.join(str(prev), file_), tmpdir / file_, create_dirs=True)
    return tmpdir


params = get_test_parameters(count=count)


@pytest.fixture(scope="session", params=params["out_var_name"])
def out_var_name(request):
    param = request.param
    return param


@pytest.fixture(scope="session", params=params["var_name"])
def var_name(request):
    param = request.param
    return param


def pytest_generate_tests(metafunc):

    if "gid" in metafunc.fixturenames:
        metafunc.parametrize("gid", params["gid"])
    # elif "var_name" in metafunc.fixturenames:
    #     metafunc.parametrize("var_name", params["var_name"])
    elif "in_var_name" in metafunc.fixturenames:
        metafunc.parametrize("in_var_name", params["in_var_name"])
    # elif "out_var_name" in metafunc.fixturenames:
    #     metafunc.parametrize("out_var_name", params["out_var_name"])



#! /usr/bin/env python
import importlib
import os
import pathlib
import re
import sys
import tempfile
from functools import partial

import click
import pkg_resources
from model_metadata import MetadataNotFoundError
from model_metadata.api import query, stage
from pytest import ExitCode
from model_metadata.scripting import as_cwd

from . import __version__
from .api import check_bmi

out = partial(click.secho, bold=True, err=True)
err = partial(click.secho, fg="red", err=True)


def validate_entry_point(ctx, param, value):
    MODULE_REGEX = r"^(?!.*\.\.)(?!.*\.$)[A-Za-z][\w\.]*$"
    CLASS_REGEX = r"^[_a-zA-Z][_a-zA-Z0-9]+$"
    if value is not None:
        try:
            module_name, class_name = value.split(":")
        except ValueError:
            raise click.BadParameter(
                "Bad entry point", param=value, param_hint="module_name:ClassName"
            )
        if not re.match(MODULE_REGEX, module_name):
            raise click.BadParameter(
                "Bad module name ({0})".format(module_name),
                param_hint="module_name:ClassName",
            )
        if not re.match(CLASS_REGEX, class_name):
            raise click.BadParameter(
                "Bad class name ({0})".format(class_name),
                param_hint="module_name:ClassName",
            )
    return value


def load_component(entry_point):
    module_name, cls_name = entry_point.split(":")

    component = None
    try:
        module = importlib.import_module(module_name)
    except ImportError:
        raise
    else:
        try:
            component = module.__dict__[cls_name]
        #     component = module.__dict__[cls_name].__name__
        except KeyError:
            raise ImportError(cls_name)

    return component


def _stage_component(entry_point, stage_dir="."):
    config_file = query(entry_point, "run.config_file.path")
    manifest = stage(entry_point, str(stage_dir))

    return config_file, manifest


def _tree(files):
    tree = []
    prefix = ["|--"] * (len(files) - 1) + ["`--"]
    for p, fname in zip(prefix, files):
        tree.append(f"{p} {fname}")
    return os.linesep.join(tree)


@click.command(
    context_settings=dict(ignore_unknown_options=True, allow_extra_args=True)
)
@click.version_option(version=__version__)
@click.option(
    "-q",
    "--quiet",
    is_flag=True,
    help=(
        "Don't emit non-error messages to stderr. Errors are still emitted, "
        "silence those with 2>/dev/null."
    ),
)
@click.option(
    "-v", "--verbose", is_flag=True, help="Also emit status messages to stderr."
)
@click.option("--help-pytest", is_flag=True, help="Print help about pytest.")
@click.option(
    "--root-dir",
    type=click.Path(
        exists=True, file_okay=False, dir_okay=True, writable=True, resolve_path=True
    ),
    help="Define root directory for BMI tests",
)
@click.option(
    "--config-file",
    type=click.Path(
        exists=True, file_okay=True, dir_okay=False, writable=True, resolve_path=True
    ),
    help="Name of model configuration file",
)
@click.option(
    "--manifest",
    type=click.Path(
        exists=True, file_okay=True, dir_okay=False, readable=True, resolve_path=True
    ),
    help="Path to manifest file of staged model input files.",
)
@click.option("--bmi-version", default="2.0", help="BMI version to test against")
@click.argument("entry_point", callback=validate_entry_point)
@click.argument("pytest_args", nargs=-1, type=click.UNPROCESSED)
def main(
    entry_point,
    root_dir,
    bmi_version,
    config_file,
    manifest,
    quiet,
    verbose,
    pytest_args,
    help_pytest,
):
    """Validate a BMI implementation.

    \b
    Examples:

      Test a BMI for the class *Hydrotrend* in module *pymt_hydrotrend*,

        $ bmi-test pymt_hydrotrend:Hydrotrend

    This will test the BMI with a default set of input files as obtained
    through the model metadata associated with the component.

    If the component you would like to test does not have model metadata
    that bmi-tester recognizes, or you would like to test with a non-default
    set of input files, use the *--root-dir* and *--config-file* options.

        $ bmi-tests pymt_hydrotrend:Hydrotrend --root-dir=my_files/ --config-file=config.txt

    where *my_files* is a folder that contains the input files to test with
    and *config.txt* is the configuration file, which will be passed to the
    *initialize* method.
    """
    if root_dir and not config_file:
        err("using --root-dir but no config file specified (use --config-file)")
        raise click.Abort()

    module_name, class_name = entry_point.split(":")

    try:
        Bmi = load_component(entry_point)
    except ImportError:
        err(f"unable to import BMI implementation, {class_name}, from {module_name}")
        raise click.Abort()

    if root_dir:
        stage_dir = root_dir
        if manifest is None:
            manifest = os.listdir(stage_dir)
    else:
        stage_dir = tempfile.mkdtemp()
        try:
            config_file, manifest = _stage_component(entry_point, stage_dir)
        except MetadataNotFoundError:
            config_file, manifest = _stage_component(class_name, stage_dir)

    stages = sorted(
        [pathlib.Path(pkg_resources.resource_filename(__name__, "bootstrap"))]
        + list(
            pathlib.Path(pkg_resources.resource_filename(__name__, "tests")).glob(
                "stage_*"
            )
        )
    )

    if not quiet:
        out("Location of tests:")
        for stage_path in (str(stage_path) for stage_path in stages):
            out(f"- {stage_path}")
        out(f"Entry point: {entry_point}")
        out(repr(Bmi()))
        out(f"BMI version: {bmi_version}")
        out(f"Stage folder: {stage_dir}")
        out(f"> tree -d {stage_dir}")
        if manifest:
            out(_tree(manifest))
        out(f"> cat {stage_dir}/{config_file}")
        with open(os.path.join(stage_dir, config_file), "r") as fp:
            out(fp.read())

    with as_cwd(stage_dir):
        for stage_path in sorted(stages):
            status = check_bmi(
                entry_point,
                tests_dir=str(stage_path),
                # tests_dir=tests_dir,
                input_file=config_file,
                manifest=manifest,
                bmi_version=bmi_version,
                extra_args=pytest_args + ("-vvv",),
                help_pytest=help_pytest,
            )
            if status != ExitCode.OK:
                break

    if not quiet:
        if status == ExitCode.OK:
            out("🎉 All tests passed!")
        else:
            err("😞 There were errors")

    sys.exit(status)
